/*
    Spakes - a client/server Snakes-clone for up to 4 players
    Copyright (C) 2001, 2002, 2004 Markus Triska

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact me at triska@gmx.at
*/

#include <algorithm>
#include <vector>
#include <queue>

#include "player.h"
#include "event.h"
#include "protocol.h"
#ifdef WINDOWS
#include "window.h"
#else
#include "gameboard.h"
#endif
#include "udpsock.h"
#include "mylog.h"

using namespace std;

extern MyLogFile logfile;

extern vector<Point> apples;

void Player::draw(bool fullredraw) { w->draw_player(this, fullredraw); }
void Player::undraw() { w->undraw_player(this); }
void Player::announce() { w->announce_player(this); }
void Player::undraw_begin() { w->undraw_begin(this); }

GameField::GameField()
{
	for (int i = 0; i < BOARDXSIZE; i++) {
		for (int j = 0; j < BOARDYSIZE; j++) {
			field[i][j] = 0;
		}
	}
}

bool GameField::free(const Point &p)
{
	return (inside(p) && field[p.x][p.y] == 0);
}

bool GameField::inside(const Point &p)
{
	return ((p.x >= 0) && (p.x < BOARDXSIZE) && (p.y >= 0) && (p.y < BOARDYSIZE));
}

void GameField::add_dot(const Point &p)
{
	if (inside(p)) field[p.x][p.y] += 1;
}

void GameField::remove_dot(const Point &p)
{
	if (inside(p)) field[p.x][p.y] -= 1;
}



GameField field;

#ifdef WINDOWS
Player::Player(Window *pw, int pplayerid, istringstream &command)
#else
Player::Player(Gameboard *pw, int pplayerid, istringstream &command)
#endif
{
	w = pw; numticks = 0; score = 0;
	numlives = 3; ingame = true; playerid = pplayerid;
	int com_x, com_y;
	command >> name >> com_x >> com_y;
	currpos = Point(com_x, com_y);
	field.add_dot(currpos);
	dots.push_back(currpos);
}


bool Player::outside()
{
	return (!field.inside(currpos));
}


void Player::lose_life()
{
	if (numlives > 0) numlives--;
	undraw();
	for (list<Point>::iterator it = dots.begin(); it != dots.end(); it++) {
		field.remove_dot(*it);
	}
	dots.clear();
}

Player::~Player()
{
	lose_life(); // Remove us from the board
}


void Player::process_event(const string &s)
{
	istringstream command(s);
	unsigned char ccom_command;
	int com_playerid;
	command >> ccom_command;
	int com_x, com_y;

	switch(ccom_command) {
	case SRV_TICK:
		if (!Living()) return;
		command >> com_playerid;
		if (com_playerid == playerid) {
			command >> com_x >> com_y;
			currpos = Point(com_x, com_y);
			field.add_dot(currpos);
			dots.push_back(currpos);
			tick();
			draw();
		}
		break;

	case SRV_HITWALL:
		if (!Living()) return;
		command >> com_playerid;
		if (com_playerid == playerid) lose_life();
		break;

	case SRV_ATEAPPLE:
		if (!Living()) return;
		command >> com_playerid;
		if (com_playerid == playerid) {
			score += 10;
			int posx, posy, newposx, newposy;
			command >> posx >> posy >> newposx >> newposy;
			Point oldpos(posx, posy);
			Point newpos(newposx, newposy);
			w->draw_apple(newpos);
			replace(apples.begin(), apples.end(), oldpos, newpos);
		}
		break;

	case SRV_PLAYER_OUT:
		command >> com_playerid;
		if (com_playerid == playerid) {
			logfile << "Player " << playerid << " quit the game\n";
			ingame = false;
			lose_life();
			numlives = 0;
		}
		break;
	}
}

void Player::tick()
{
	if (!Living()) return;
	numticks++;
	if ((dots.size() > 1) && (numticks % 2 == 0)) {
		undraw_begin();
		Point begin = dots.front();
		field.remove_dot(begin);
		dots.pop_front();
	}
}


bool Player::collision()
{
	return field.fieldval(currpos) > 1;
}

bool operator==(const ASNode &n1, const ASNode &n2) {
	return n1.pos == n2.pos;
}


#ifdef WINDOWS
AIPlayer::AIPlayer(Window *pw, int pplayerid, std::istringstream &s) : Player(pw, pplayerid, s)
#else
AIPlayer::AIPlayer(Gameboard *pw, int pplayerid, std::istringstream &s) : Player(pw, pplayerid, s)
#endif
{
	startingpos = currpos;
}

void AIPlayer::push_successors(Point &to, ASNode node)
{
	for (int i = -1; i < 2; i++) {
		for (int j = -1; j < 2; j++) {
 			if ((i == 0) && (j == 0)) continue;
 			if ((i == j) || (i == -j)) continue;
			Point nextpos(node.pos.x + i, node.pos.y + j);

			if (field.free(nextpos)) {
				ASNode newnode(nextpos);
				newnode.path = node.path;
				newnode.path.push_back(nextpos);

				int heur = newnode.path.size() + nextpos.dist(to);

				newnode.fitness = -heur;

				nodes_it cl = find(closedlist.begin(), closedlist.end(), newnode);
				if (cl != closedlist.end()) {
					if (cl->fitness >= newnode.fitness) {
						continue;
					} else {
						closedlist.erase(cl);
					}
				}

				nodes_it ol = find(openlist.begin(), openlist.end(), newnode);

				if (ol != openlist.end()) {
					if (ol->fitness >= newnode.fitness) {
						continue;
					} else {
						openlist.erase(ol);
					}
				}
				openlist.push_back(newnode);
			}
		}
	}
}


void AIPlayer::clear_lists()
{
	openlist.clear();
	closedlist.clear();
	path.clear();
}

bool AIPlayer::compute_route(Point &from, Point &to)
{
	// A* search

	logfile << "I want to go from (" << from.x << "/" << from.y << ") to (" << to.x << "/" << to.y << ")\n";

	clear_lists();

	ASNode pfrom(from);
	openlist.push_back(pfrom);

	while (!openlist.empty()) {
		nodes_it bestit = openlist.begin();

		for (nodes_it it = openlist.begin(); it != openlist.end(); it++) {
			if (it->pos == to) {
				path = it->path;
				return true;
			}

			if (it->fitness > bestit->fitness) bestit = it;
		}

		ASNode n = *bestit;
		openlist.erase(bestit);

		push_successors(to, n);
		closedlist.push_back(n);
	}

	return false;
}

void AIPlayer::lose_life()
{
	Player::lose_life();
	clear_lists();
	currpos = startingpos;
	path.clear();
	if (Living()) {
		dots.push_back(currpos);
		field.add_dot(currpos);
	}
}

typedef vector<Point> points;
typedef points::iterator points_it;

#define PUSHSUCC(a) if (field.free(a)) {\
   if (floodfield[(a).x][(a).y] == apple) {\
       headingfor = a; return true;\
   } else if (floodfield[(a).x][(a).y] == unvisited) {\
       floodfield[(a).x][(a).y] = visited;\
       floodlist.push_back(a);\
   }\
}


bool AIPlayer::choose_target()
{
	// floodfill to see which apple can be reached at all

	logfile << "I'm choosing a new target\n";

	const int unvisited = 0;
	const int apple = 1;
	const int visited = 2;

	int floodfield[BOARDXSIZE][BOARDYSIZE];
	for (int i = 0; i < BOARDXSIZE; i++) {
		for (int j = 0; j < BOARDYSIZE; j++) floodfield[i][j] = unvisited;
	}


	for (points_it it = apples.begin(); it != apples.end(); it++) {
		floodfield[it->x][it->y] = apple;
	}


	list<Point> floodlist;

	floodlist.push_back(currpos);
	floodfield[currpos.x][currpos.y] = visited;

	while (!floodlist.empty()) {
		Point current = floodlist.front();
		floodlist.pop_front();

		PUSHSUCC(current + Point(-1,0))
		PUSHSUCC(current + Point(0,-1))
		PUSHSUCC(current + Point(1,0))
		PUSHSUCC(current + Point(0,1))
	}

	// no apple can be reached
	return false;

}

void AIPlayer::tick()
{
	if (!Living()) return;
	Player::tick();

	for (list<Point>::const_iterator it = path.begin(); it != path.end(); it++) {
		if (!field.free(*it)) {
			path.clear();
			break;
		}
	}

	if (path.empty() && choose_target()) {
		compute_route(currpos, headingfor);
	}

	if (path.empty()) { // STILL empty => move on
		currpos = currpos + lastdirection;
	} else {
		Point nextpos = path.front();
		path.pop_front();
		lastdirection = nextpos - currpos;
		currpos = nextpos;
	}

	field.add_dot(currpos);
	dots.push_back(currpos);
	draw();
}

#ifdef WINDOWS
Consoleplayer::Consoleplayer(Window *pw, int pplayerid, istringstream &s, bool pusearrows) : Player(pw, pplayerid, s)
#else
Consoleplayer::Consoleplayer(Gameboard *pw, int pplayerid, istringstream &s, bool pusearrows) : Player(pw, pplayerid, s)
#endif
{
	s >> startingdirection;
	usearrows = pusearrows;
	direction = startingdirection;
	startingpos = currpos;
	for (int i = 0; i < 4; i++) keypressed[i] = false;
}


/* should make these constants available in server/init.cpp */
const int DIR_UP = 1, DIR_UPRIGHT = 2, DIR_RIGHT = 3, DIR_DOWNRIGHT = 4,
	DIR_DOWN = 5, DIR_DOWNLEFT = 6, DIR_LEFT = 7, DIR_UPLEFT = 8;

void Consoleplayer::process_event(const Event &e)
{
	if (usearrows) {
		switch(e.type) {
		case KEYUP: keypressed[e.value1] = false; break;
		case KEYDOWN: keypressed[e.value1] = true;
		}

		if (keypressed[UPARROW] && keypressed[RIGHTARROW]) {
			if (direction != DIR_DOWNLEFT) direction = DIR_UPRIGHT;
		} else if (keypressed[RIGHTARROW] && keypressed[DOWNARROW]) {
			if (direction != DIR_UPLEFT) direction = DIR_DOWNRIGHT;
		} else if (keypressed[DOWNARROW] && keypressed[LEFTARROW]) {
			if (direction != DIR_UPRIGHT) direction = DIR_DOWNLEFT;
		} else if (keypressed[LEFTARROW] && keypressed[UPARROW]) {
			if (direction != DIR_DOWNRIGHT) direction = DIR_UPLEFT;
		} else if (keypressed[UPARROW]) {
			if (direction != DIR_DOWN) direction = DIR_UP;
		} else if (keypressed[RIGHTARROW]) {
			if (direction != DIR_LEFT) direction = DIR_RIGHT;
		} else if (keypressed[DOWNARROW]) {
			if (direction != DIR_UP) direction = DIR_DOWN;
		} else if (keypressed[LEFTARROW]) {
			if (direction != DIR_RIGHT) direction = DIR_LEFT;
		}
	} else {
		switch(e.value1) {
		case UPARROW:
			if (direction != DIR_DOWN) direction = DIR_UP;
			break;
		case UPRIGHTARROW:
			if (direction != DIR_DOWNLEFT) direction = DIR_UPRIGHT;
			break;
		case RIGHTARROW:
			if (direction != DIR_LEFT) direction = DIR_RIGHT;
			break;
		case DOWNRIGHTARROW:
			if (direction != DIR_UPLEFT) direction = DIR_DOWNRIGHT;
			break;
		case DOWNARROW:
			if (direction != DIR_UP) direction = DIR_DOWN;
			break;
		case DOWNLEFTARROW:
			if (direction != DIR_UPRIGHT) direction = DIR_DOWNLEFT;
			break;
		case LEFTARROW:
			if (direction != DIR_RIGHT) direction = DIR_LEFT;
			break;
		case UPLEFTARROW:
			if (direction != DIR_DOWNRIGHT) direction = DIR_UPLEFT;
			break;
		}
	}
}

void Consoleplayer::tick()
{
	if (!Living()) return;
	Player::tick();
	switch (direction) {
	case DIR_UP: currpos.y -= 1; break;
	case DIR_UPRIGHT: currpos.x += 1; currpos.y -= 1; break;
	case DIR_RIGHT: currpos.x += 1; break;
	case DIR_DOWNRIGHT: currpos.x += 1; currpos.y += 1; break;
	case DIR_DOWN: currpos.y += 1; break;
	case DIR_DOWNLEFT: currpos.x -= 1; currpos.y += 1; break;
	case DIR_LEFT: currpos.x -= 1; break;
	case DIR_UPLEFT: currpos.x -= 1; currpos.y -= 1; break;
	}
	dots.push_back(currpos);
	field.add_dot(currpos);
	draw();
}

void Consoleplayer::lose_life()
{
	Player::lose_life();
	direction = startingdirection;
	currpos = startingpos;
	if (Living()) {
		dots.push_back(currpos);
		field.add_dot(currpos);
	}
}



#ifdef WINDOWS
RemotePrimaryPlayer::RemotePrimaryPlayer(Window *pw, int pplayerid, istringstream &command) : Player(pw, pplayerid, command)
#else
RemotePrimaryPlayer::RemotePrimaryPlayer(Gameboard *pw, int pplayerid, istringstream &command) : Player(pw, pplayerid, command)
#endif
{
	num_scheduled_ticks = 0;
	scheduled_ticks = new ostringstream;
	string ip;
	unsigned short port;
	command >> ip >> port;
	socket = new UDPSocket(ip, port);
}

void RemotePrimaryPlayer::send_tick(Player *p)
{
	if (p->playerid == playerid) return;
	if (!ingame) return;

	/* Borland C++ 5.5 has problems with ostringstream.clear(),
	   so I (re-)allocate dynamically to clear it */
	*scheduled_ticks << SRV_TICK << p->playerid << ' ' << p->currpos.x << ' ' << p->currpos.y << DELIMINATER;

	num_scheduled_ticks++;

	if (num_scheduled_ticks >= 2) {
		num_scheduled_ticks = 0;
		string tosend = (*scheduled_ticks).str();
 		socket->Send(tosend);
		delete scheduled_ticks;
		scheduled_ticks = new ostringstream;
	}
}

RemotePrimaryPlayer::~RemotePrimaryPlayer()
{
	delete socket;
	delete scheduled_ticks;
}


